#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import logging
import os

import tentacles_manager.tentacle_package_util as TentaclePackageUtil
import tentacles_manager.tentacle_util as TentacleUtil

from tentacles_manager import TENTACLE_PACKAGE_DESCRIPTION, EVALUATOR_DEFAULT_FOLDER, \
    TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION, \
    TENTACLE_DESCRIPTION_IS_URL, TENTACLE_TYPES, EVALUATOR_CONFIG_FOLDER, TENTACLE_MODULE_NAME, TENTACLE_MODULE_TYPE, \
    TENTACLE_MODULE_SUBTYPE, TENTACLE_MODULE_VERSION, TENTACLE_MODULE_CONFIG_FILES, TENTACLE_MODULE_REQUIREMENTS, \
    TENTACLE_MODULE_REQUIREMENT_WITH_VERSION, TENTACLES_PATH, PYTHON_INIT_FILE, TENTACLE_MODULE_TESTS, \
    TentacleManagerActions, CONFIG_DEFAULT_EVALUATOR_FILE, CONFIG_EVALUATOR_FILE_PATH, TENTACLE_MODULE_DEV, \
    TENTACLE_PACKAGE_NAME, TENTACLE_MODULE_RESOURCE_FILES, EVALUATOR_RESOURCE_FOLDER, CONFIG_TRADING_FILE_PATH, \
    CONFIG_DEFAULT_TRADING_FILE, TENTACLE_MODULE_CONFIG_SCHEMA_FILES, INFO


class TentaclePackageManager:
    def __init__(self, config, tentacle_manager):
        self.config = config
        self.tentacle_manager = tentacle_manager
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(INFO)
        self.just_processed_modules = []
        self.installed_modules = {}
        self.max_steps = None
        self.current_step = 1

    def _process_action_on_module(self, action, module_type, module_subtype,
                                  module_version, module_file_content, module_test_files, target_folder, module_name):

        if module_subtype in TENTACLE_TYPES and (module_subtype or module_subtype in TENTACLE_TYPES):

            did_something = False

            # Update module file
            module_file_dir = TentacleUtil.create_path_from_type(module_type, module_subtype, target_folder)

            module_file_path = f"{module_file_dir}/{module_name}.py"

            # Write the new file in locations
            if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
                # Install package in evaluator
                with open(module_file_path, "w") as module_file:
                    module_file.write(module_file_content)
                    did_something = True

            # Remove package line from init file
            elif action == TentacleManagerActions.UNINSTALL:
                try:
                    if os.path.exists(module_file_path):
                        os.remove(module_file_path)
                        did_something = True
                except OSError:
                    did_something = False

            # Update local __init__
            lines_in_init = f"""try:
    check_tentacle(PATH, '{module_name}')
    from .{module_name} import *
except Exception as e:
    LOGGER.error(f'Error when loading {module_name}: {{e}}')
"""
            init_file = f"{TENTACLES_PATH}/{TENTACLE_TYPES[module_type]}/{TENTACLE_TYPES[module_subtype]}/" \
                f"{target_folder}/{PYTHON_INIT_FILE}"

            self.update_init_file(action, init_file, lines_in_init)

            # Update module test files
            test_file_dir = TentacleUtil.create_path_from_type(module_type, module_subtype, target_folder, True)
            for test_file, test_file_content in module_test_files.items():
                module_test_file_path = f"{test_file_dir}/{test_file}.py"

                # Write the new file in locations
                if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
                    # Install package in evaluator
                    with open(module_test_file_path, "w") as installed_module_test_file:
                        installed_module_test_file.write(test_file_content)

                # Remove package line from init file
                elif action == TentacleManagerActions.UNINSTALL:
                    try:
                        os.remove(module_test_file_path)
                    except OSError:
                        pass

            if action == TentacleManagerActions.INSTALL:
                self.logger.info(f"{self._format_current_step()}{module_name} {module_version} "
                                 f"successfully installed in: {module_file_dir}")
            elif action == TentacleManagerActions.UNINSTALL:
                if did_something:
                    self.logger.info(f"{self._format_current_step()}{module_name} {module_version} "
                                     f"successfully uninstalled (file: {module_file_dir})")
            elif action == TentacleManagerActions.UPDATE:
                self.logger.info(f"{self._format_current_step()}{module_name} successfully updated to version "
                                 f"{module_version} in: {module_file_dir}")
            self.just_processed_modules.append(TentacleUtil.get_full_module_identifier(module_name, module_version))
            return did_something

        else:
            raise Exception("Tentacle type not found")

    def process_module(self, action, package, module_name, package_localisation, is_url, target_folder, package_name):
        parsed_module = TentacleUtil.parse_module_header(package[module_name])
        module_type = parsed_module[TENTACLE_MODULE_TYPE]
        module_subtype = parsed_module[TENTACLE_MODULE_SUBTYPE]
        module_tests = parsed_module[TENTACLE_MODULE_TESTS]
        module_file_content = ""
        module_dev = parsed_module[TENTACLE_MODULE_DEV]
        module_test_files = {test: "" for test in module_tests} if module_tests else {}

        if TentacleUtil.install_on_development(self.config, module_dev):
            if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
                location_str = TentacleUtil.create_localization_from_type(package_localisation,
                                                                          module_type,
                                                                          module_subtype,
                                                                          module_name)
                module_loc = f"{location_str}.py"

                if is_url:
                    module_file_content = TentaclePackageUtil.get_package_file_content_from_url(module_loc)
                else:
                    with open(module_loc, "r") as module_file:
                        module_file_content = module_file.read()

                module_file_content = TentaclePackageUtil.add_package_name(module_file_content, package_name)

                if module_test_files:
                    for test in module_tests:
                        location_str = TentacleUtil.create_localization_from_type(package_localisation,
                                                                                  module_type,
                                                                                  module_subtype,
                                                                                  test,
                                                                                  True)
                        test_loc = f"{location_str}.py"

                        if is_url:
                            module_test_files[test] = TentaclePackageUtil.get_package_file_content_from_url(test_loc)
                        else:
                            with open(test_loc, "r") as module_file:
                                module_test_files[test] = module_file.read()

            if self._process_action_on_module(action, module_type, module_subtype,
                                              parsed_module[TENTACLE_MODULE_VERSION],
                                              module_file_content, module_test_files, target_folder, module_name):
                # manage module config
                self._try_action_on_config_or_resources(action, parsed_module, module_name, is_url, package_localisation)

            if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
                self._try_action_on_requirements(action, parsed_module, package_name)
        elif action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
            self.logger.info(f"{module_name} is currently on development, "
                             "it will not be installed (to install it anyway, "
                             "add \"DEV-MODE\": true in your config.json)")

    def _should_do_something(self, action, module_name, module_version, need_this_exact_version=False, requiring=None,
                             parsed_module=None, package=None):
        if action == TentacleManagerActions.UPDATE:
            if TentacleUtil.is_module_in_list(module_name, None, self.installed_modules):
                installed_version = self.installed_modules[module_name][TENTACLE_MODULE_VERSION]
                if not need_this_exact_version:
                    new_version_available = TentacleUtil.is_first_version_superior(module_version, installed_version)
                    if not new_version_available:
                        self.logger.info(f"{self._format_current_step()}{module_name} version {installed_version} "
                                         f"is already up to date")
                    return new_version_available
                else:
                    is_required_version_installed = True
                    print_version = "any version"
                    if module_version is not None:
                        is_required_version_installed = installed_version == module_version
                        print_version = f"version {module_version}"
                    if is_required_version_installed:
                        self.logger.info(f"{requiring} dependency: {module_name} {print_version} is satisfied.")
                        return False
                    else:
                        return True
            else:
                if requiring:
                    self.logger.error(f"can't find tentacle: {module_name} required for {requiring} in installed "
                                      f"tentacles. Try to install the required tentacle")
                else:
                    if parsed_module is None:
                        parsed_module = TentacleUtil.parse_module_header(package)
                    if TentacleUtil.install_on_development(self.config, parsed_module[TENTACLE_MODULE_DEV]):
                        self.logger.info(f"new tentacle found in tentacles packages: {module_name}. "
                                         f"You can install it using the command: start.py -p install {module_name}")

            return False
        else:
            return True

    def try_action_on_tentacles_package(self, action, package, target_folder):
        package_description = package[TENTACLE_PACKAGE_DESCRIPTION]
        package_localisation = package_description[TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION]
        is_url = package_description[TENTACLE_DESCRIPTION_IS_URL]
        package_name = package_description[TENTACLE_PACKAGE_NAME]
        for tentacle_module in package:
            try:
                if tentacle_module != TENTACLE_PACKAGE_DESCRIPTION and \
                        not self._has_just_processed_module(tentacle_module,
                                                            package[tentacle_module][TENTACLE_MODULE_VERSION]) and \
                        self._should_do_something(action, tentacle_module,
                                                  package[tentacle_module][TENTACLE_MODULE_VERSION],
                                                  package=package[tentacle_module]):
                    self.process_module(action, package, tentacle_module, package_localisation, is_url, target_folder,
                                        package_name)
            except Exception as e:
                error = f"failed for tentacle module '{tentacle_module}' ({e})"
                if action == TentacleManagerActions.INSTALL:
                    self.logger.error(f"Installation {error}")
                elif action == TentacleManagerActions.UNINSTALL:
                    self.logger.error(f"Uninstalling {error}")
                elif action == TentacleManagerActions.UPDATE:
                    self.logger.error(f"Updating {error}")
            self.inc_current_step()

    def _try_action_on_requirements(self, action, parsed_module, package_name):
        success = True
        module_name = parsed_module[TENTACLE_MODULE_NAME]
        applied_modules = [module_name]
        if parsed_module[TENTACLE_MODULE_REQUIREMENTS]:
            for requirement_data in parsed_module[TENTACLE_MODULE_REQUIREMENTS]:
                if success:
                    requirement_module_name = requirement_data[TENTACLE_MODULE_NAME]
                    requirement_module_version = requirement_data[TENTACLE_MODULE_VERSION]
                    if not self._has_just_processed_module(requirement_module_name, requirement_module_version) and \
                            self._should_do_something(action, requirement_module_name,
                                                      requirement_module_version, True, module_name,
                                                      parsed_module=parsed_module):
                        try:
                            req_package, _, localisation, is_url, destination, _ = self.tentacle_manager. \
                                get_package_in_lists(requirement_module_name, requirement_module_version)

                            if req_package:
                                self.process_module(action, req_package, requirement_module_name,
                                                    localisation, is_url, destination, package_name)
                                applied_modules.append(requirement_module_name)
                            else:
                                raise Exception(f"module requirement '"
                                                f"{requirement_data[TENTACLE_MODULE_REQUIREMENT_WITH_VERSION]}' "
                                                f"not found in package lists")

                        except Exception as e:
                            error = f"failed for tentacle module requirement '{requirement_module_name}' " \
                                f"of module {module_name} ({e})"
                            if action == TentacleManagerActions.INSTALL:
                                self.logger.error(f"installation {error}")
                            elif action == TentacleManagerActions.UNINSTALL:
                                self.logger.error(f"uninstalling {error}")
                            elif action == TentacleManagerActions.UPDATE:
                                self.logger.error(f"updating {error}")
                            success = False

            # failed to install requirements
            if not success:
                if action == TentacleManagerActions.UPDATE:
                    # uninstall module
                    req_package, _, localisation, is_url, destination, _ = self.tentacle_manager. \
                        get_package_in_lists(module_name)
                    if req_package:
                        self.process_module(TentacleManagerActions.UNINSTALL, req_package, module_name,
                                            localisation, is_url, destination, package_name)

                elif action == TentacleManagerActions.INSTALL:
                    # uninstall module and requirements
                    for module_to_remove in applied_modules:
                        req_package, _, localisation, is_url, destination, _ = self.tentacle_manager. \
                            get_package_in_lists(module_to_remove)
                        if req_package:
                            self.process_module(TentacleManagerActions.UNINSTALL, req_package, module_to_remove,
                                                localisation, is_url, destination, package_name)

    def _try_action_on_config_or_resources(self, action, parsed_module, module_name, is_url, package_localisation):

        file_dir = TentacleUtil.create_path_from_type(parsed_module[TENTACLE_MODULE_TYPE],
                                                      parsed_module[TENTACLE_MODULE_SUBTYPE], "")

        if parsed_module[TENTACLE_MODULE_CONFIG_FILES]:
            for config_file in parsed_module[TENTACLE_MODULE_CONFIG_FILES]:
                config_file_path = f"{file_dir}{EVALUATOR_CONFIG_FOLDER}/{config_file}"
                default_config_file_path = \
                    f"{file_dir}{EVALUATOR_CONFIG_FOLDER}/{EVALUATOR_DEFAULT_FOLDER}/{config_file}"

                self._try_action_on_config_or_resource_file(action,
                                                            module_name,
                                                            package_localisation,
                                                            is_url,
                                                            parsed_module,
                                                            config_file,
                                                            config_file_path,
                                                            default_file_path=default_config_file_path)

        if parsed_module[TENTACLE_MODULE_CONFIG_SCHEMA_FILES]:
            for config_schema_file in parsed_module[TENTACLE_MODULE_CONFIG_SCHEMA_FILES]:
                config_schema_file_path = f"{file_dir}{EVALUATOR_CONFIG_FOLDER}/{config_schema_file}"
                default_config_schema_file_path = \
                    f"{file_dir}{EVALUATOR_CONFIG_FOLDER}/{EVALUATOR_DEFAULT_FOLDER}/{config_schema_file}"

                self._try_action_on_config_or_resource_file(action,
                                                            module_name,
                                                            package_localisation,
                                                            is_url,
                                                            parsed_module,
                                                            config_schema_file,
                                                            config_schema_file_path,
                                                            default_file_path=default_config_schema_file_path,
                                                            force_refresh=True)

        if parsed_module[TENTACLE_MODULE_RESOURCE_FILES]:
            for res_file in parsed_module[TENTACLE_MODULE_RESOURCE_FILES]:
                res_file_path = f"{file_dir}{EVALUATOR_RESOURCE_FOLDER}/{res_file}"

                self._try_action_on_config_or_resource_file(action,
                                                            module_name,
                                                            package_localisation,
                                                            is_url,
                                                            parsed_module,
                                                            res_file,
                                                            res_file_path,
                                                            read_as_bytes=True)

    def _try_action_on_config_or_resource_file(self, action,
                                               module_name,
                                               package_localisation,
                                               is_url,
                                               parsed_module,
                                               file,
                                               file_path,
                                               default_file_path=None,
                                               read_as_bytes=False,
                                               force_refresh=False):

        if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:

            try:
                # get config file content from localization
                module_loc = TentacleUtil.create_localization_from_type(package_localisation,
                                                                        parsed_module[TENTACLE_MODULE_TYPE],
                                                                        parsed_module[TENTACLE_MODULE_SUBTYPE],
                                                                        file)

                if is_url:
                    config_file_content = TentaclePackageUtil.get_package_file_content_from_url(module_loc,
                                                                                                as_bytes=read_as_bytes)
                else:
                    with open(module_loc, "rb" if read_as_bytes else "r") as module_file:
                        config_file_content = module_file.read()

                if force_refresh:
                    with open(file_path, "wb" if read_as_bytes else "w") as new_file:
                        new_file.write(config_file_content)
                else:
                    # install local file content
                    if action == TentacleManagerActions.INSTALL and \
                            TentaclePackageUtil.should_recreate_config_file(file_path, config_file_content):
                        with open(file_path, "wb" if read_as_bytes else "w") as new_file:
                            new_file.write(config_file_content)

                    # copy into default
                    if default_file_path:
                        with open(default_file_path, "wb" if read_as_bytes else "w") as new_default_file:
                            new_default_file.write(config_file_content)

                        if action == TentacleManagerActions.UPDATE:
                            self.logger.info(f"{file} configuration / resource file for {module_name} module ignored "
                                             f"to save the current configuration. The default configuration file has "
                                             f"been updated in: {default_file_path}.")

            except Exception as e:
                raise Exception(f"Fail to install configuration / resource : {e}")

        elif action == TentacleManagerActions.UNINSTALL:
            try:
                os.remove(file_path)

                if default_file_path:
                    os.remove(default_file_path)
            except OSError:
                pass

    def _has_just_processed_module(self, module_name, module_version):
        return TentacleUtil.is_module_in_list(module_name, module_version, self.just_processed_modules)

    @staticmethod
    def get_installed_modules():
        modules = {}
        # Foreach folder (not hidden)
        for root_dir in os.listdir(os.getcwd()):
            if os.path.isdir(root_dir) and not root_dir.startswith('.'):
                TentaclePackageUtil.read_tentacles(root_dir, modules)
        return modules

    def _format_current_step(self):
        return f"{self.current_step}/{self.max_steps}: " if self.max_steps is not None else ""

    def set_max_steps(self, max_steps):
        self.max_steps = max_steps

    def set_installed_modules(self, installed_modules):
        self.installed_modules = installed_modules

    def inc_current_step(self):
        self.current_step += 1

    @staticmethod
    def update_init_file(action, init_file, lines_in_init):
        init_content = ""
        init_file_exists = os.path.isfile(init_file)
        if init_file_exists:
            with open(init_file, "r") as init_file_r:
                init_content = init_file_r.read()
        if not init_file_exists or not init_content:
            init_content = TentaclePackageManager._create_init_file(init_file)

        # check if line already exists
        line_exists = False if init_content.find(lines_in_init) == -1 else True

        # Add new package in init file
        if action == TentacleManagerActions.INSTALL or action == TentacleManagerActions.UPDATE:
            if not line_exists:
                with open(init_file, "w") as init_file_w:
                    # add new package to init
                    init_file_w.write(init_content + lines_in_init)

        # remove package line from init file
        elif action == TentacleManagerActions.UNINSTALL:
            if line_exists:
                with open(init_file, "w") as init_file_w:
                    # remove package to uninstall from init
                    init_file_w.write(init_content.replace(lines_in_init, ""))

    @staticmethod
    def _create_init_file(init_file):
        with open(init_file, "w") as init_file_w:
            init_file_content = """import os

from tentacles_manager.tentacle_util import check_tentacle
import logging

LOGGER = logging.getLogger("TentacleLoader")
PATH = os.path.dirname(os.path.realpath(__file__))

"""
            init_file_w.write(init_file_content)
            return init_file_content

    @staticmethod
    def _log_config_file_update_exception(logger, exception):
        logger.exception(exception)
        logger.error(f"Something went wrong when checking installed tentacles: {exception}.\nIf Octobot is now "
                     "working after this, you should re-install your tentacles (start.py -p install all).\nIf this "
                     "problem keeps appearing, try to reset all tentacles (start.py -p reset_tentacles).")

    @staticmethod
    def update_evaluator_config_file(evaluator_config_file=CONFIG_EVALUATOR_FILE_PATH):

        logger = logging.getLogger(TentaclePackageManager.__name__)
        logger.setLevel(INFO)
        try:
            logger.info(f"Updating {evaluator_config_file} using new data...")

            from evaluator.RealTime import RealTimeEvaluator
            from evaluator.Social import SocialEvaluator
            from evaluator.Strategies import StrategiesEvaluator
            from evaluator.TA import TAEvaluator

            evaluators_in_config = [TAEvaluator, SocialEvaluator, RealTimeEvaluator, StrategiesEvaluator]

            TentaclePackageUtil.update_config_file(evaluator_config_file, CONFIG_DEFAULT_EVALUATOR_FILE,
                                                   evaluators_in_config)
        except Exception as e:
            TentaclePackageManager._log_config_file_update_exception(logger, e)

    @staticmethod
    def update_trading_config_file(trading_config_file=CONFIG_TRADING_FILE_PATH):

        logger = logging.getLogger(TentaclePackageManager.__name__)
        logger.setLevel(INFO)
        try:
            logger.info(f"Updating {trading_config_file} using new data...")

            from trading.trader.modes import AbstractTradingMode

            trading_modes_in_config = [AbstractTradingMode]

            TentaclePackageUtil.update_config_file(trading_config_file, CONFIG_DEFAULT_TRADING_FILE,
                                                   trading_modes_in_config)
        except Exception as e:
            TentaclePackageManager._log_config_file_update_exception(logger, e)



# TODO

import fileinput
import re
import sys
from os import remove

from tentacles_manager.tentacle_manager import TentacleManager
from tentacles_manager.tentacle_util import delete_tentacles_arch
from tentacles_manager.tentacle_package_util import read_tentacles
from tentacles_manager import TENTACLES_PATH, TENTACLES_EVALUATOR_PATH, TENTACLES_EVALUATOR_STRATEGIES_PATH, \
    EVALUATOR_DEFAULT_FOLDER, TENTACLE_MODULE_VERSION, EVALUATOR_CONFIG_FOLDER


# Warning: limit github calls with using install all as little as possible (use install momentum_evaluator instead)


def test_import():
    TentacleManager({})


def test_install_default_branch():
    manager = TentacleManager({})
    _install_test_tentacle(manager)
    _cleanup()


def test_config_file_management():
    manager = TentacleManager({})
    _install_test_tentacle(manager)
    test_tentacle = "staggered_orders_strategy_evaluator"
    tentacle_evaluator_path = f"{TENTACLES_PATH}/{TENTACLES_EVALUATOR_PATH}/{TENTACLES_EVALUATOR_STRATEGIES_PATH}"
    tentacle_path = f"{tentacle_evaluator_path}/{EVALUATOR_DEFAULT_FOLDER}"
    tentacle_name = f"{test_tentacle}.py"

    # change tentacle version to 0 to force update
    _change_tentacle_version(f"{tentacle_path}/{tentacle_name}")
    tentacle_desc = {}
    read_tentacles(tentacle_path, tentacle_desc)
    assert tentacle_desc[test_tentacle][TENTACLE_MODULE_VERSION] == "0.0.0"

    # edit tentacle configuration file
    tentacle_config_file = f"{tentacle_evaluator_path}/{EVALUATOR_CONFIG_FOLDER}" \
        f"/StaggeredOrdersStrategiesEvaluator.json"
    original_config_content = _get_file_content(tentacle_config_file)
    _change_tentacle_config_file(tentacle_config_file)
    edited_config_content = _get_file_content(tentacle_config_file)

    command_update = ["update", test_tentacle]
    manager.parse_commands(command_update, force=True)

    # ensure tentacle got updated
    read_tentacles(tentacle_path, tentacle_desc)
    assert tentacle_desc[test_tentacle][TENTACLE_MODULE_VERSION] != "0.0.0"

    # ensure tentacle configuration file did not change with update
    new_config_content = _get_file_content(tentacle_config_file)
    assert edited_config_content != original_config_content
    assert edited_config_content == new_config_content

    # now check with install command
    _install_test_tentacle(manager)

    # ensure tentacle configuration file did not change with update
    new_config_content = _get_file_content(tentacle_config_file)
    assert edited_config_content == new_config_content

    # now remove config file and ensure install recreates it
    remove(tentacle_config_file)
    _install_test_tentacle(manager)

    # ensure config file got restored
    new_config_content = _get_file_content(tentacle_config_file)
    assert original_config_content == new_config_content

    _cleanup()


def test_install_update_specified_argument_branch():
    manager = TentacleManager({})
    # use default_git_branch
    command_install = ["install", "all"]
    manager.parse_commands(command_install, force=True, default_git_branch="dev")
    manager = TentacleManager({})
    command_install = ["update", "all"]
    manager.parse_commands(command_install, force=True, default_git_branch="dev")
    _cleanup()
    # use command_install branch
    command_install = ["install", "momentum_evaluator", "branch", "dev"]
    manager.parse_commands(command_install, force=True, default_git_branch="xyz")
    _cleanup()


def test_install_dev_branch():
    manager = TentacleManager({})
    # command: install all branch dev
    command_install = ["install", "all", "branch", "dev"]
    manager.parse_commands(command_install, force=True)
    _cleanup()
    # command: install all branch=dev
    command_install = ["install", "momentum_evaluator", "branch=dev"]
    manager.parse_commands(command_install, force=True)
    _cleanup()
    # command: install all branch= dev
    command_install = ["install", "momentum_evaluator", "branch=", "dev"]
    manager.parse_commands(command_install, force=True)
    _cleanup()
    # command: install momentum_evaluator trend_evaluator branch dev
    command_install = ["install", "momentum_evaluator", "trend_evaluator", "branch", "dev"]
    manager.parse_commands(command_install, force=True)
    _cleanup()


def _cleanup():
    delete_tentacles_arch()


def _edit_file(file_path, line_pattern, replace_regex, replace_string):
    with fileinput.input(file_path, inplace=True) as f:
        for line in f:
            if line_pattern in line:
                line = re.sub(replace_regex, replace_string, line)
            sys.stdout.write(line)


def _change_tentacle_version(tentacle_path):
    _edit_file(tentacle_path, '"version": "', r"\d", "0")


def _change_tentacle_config_file(file_path):
    _edit_file(file_path, '"required_time_frames" : ', r"\[]", '["1m"]')


def _install_test_tentacle(tentacle_manager):
    command_install = ["install", "staggered_orders_strategy_evaluator"]
    tentacle_manager.parse_commands(command_install, force=True)


def _get_file_content(file_path):
    with open(file_path) as file:
        return file.read()



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import logging

import tentacles_manager.tentacle_package_util as TentaclePackageUtil
import tentacles_manager.tentacle_util as TentacleUtil
from tentacles_manager.tentacle_package_manager import TentaclePackageManager

from tentacles_manager import TENTACLE_PACKAGE_DESCRIPTION, EVALUATOR_DEFAULT_FOLDER, CONFIG_TENTACLES_KEY, \
    TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION, TENTACLE_DESCRIPTION_IS_URL, EVALUATOR_ADVANCED_FOLDER, \
    TentacleManagerActions, TENTACLE_PACKAGE_NAME, TENTACLES_DEFAULT_BRANCH, INFO


class TentacleManager:

    TENTACLE_INSTALLATION_FOUND_MESSAGE = "Tentacles installation found on this OctoBot, this action will " \
                                          "replace every local tentacle file and their configuration by their " \
                                          "remote equivalent for the command's tentacles, continue ?"

    def __init__(self, config):
        self.config = config
        self.tentacle_package_manager = TentaclePackageManager(config, self)
        self.default_package = None
        self.advanced_package_list = []
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(INFO)
        self.force_actions = False
        self.git_branch = TENTACLES_DEFAULT_BRANCH

    def install_tentacle_package(self, package_path_or_url, force=False):
        self.update_list()
        package = TentaclePackageUtil.get_package_description_with_adaptation(package_path_or_url,
                                                                              git_branch=self.git_branch)
        should_install = force
        if TentacleUtil.create_missing_tentacles_arch() and not force:
            should_install = self._confirm_action(self.TENTACLE_INSTALLATION_FOUND_MESSAGE)
        if should_install:
            self.tentacle_package_manager.set_max_steps(len(package) - 1)
            self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.INSTALL,
                                                                          package, EVALUATOR_ADVANCED_FOLDER)

    def _update_git_branch(self, commands, default_git_branch):
        branch_cmd = "branch"
        equals_separator = "="
        branch_equals_cmd = f"{branch_cmd}{equals_separator}"
        # check if specific branch is specified in commands
        # if found, set self.git_branch and remove info from commands
        if branch_cmd in commands and len(commands)-1 > commands.index(branch_cmd):
            cmd_index = commands.index(branch_cmd)
            self.git_branch = commands[cmd_index+1]
            commands.pop(cmd_index+1)
            commands.pop(cmd_index)
        elif any(branch_equals_cmd in command for command in commands):
            for command in commands:
                if branch_equals_cmd in command:
                    cmd_index = commands.index(command)
                    potential_branch = command.split(equals_separator)[1]
                    if potential_branch:
                        self.git_branch = potential_branch
                        commands.pop(cmd_index)
                        return
                    elif len(commands)-1 > commands.index(command):
                        self.git_branch = commands[commands.index(command)+1]
                        commands.pop(cmd_index+1)
                        commands.pop(cmd_index)
                        return
        else:
            # otherwise use default
            self.git_branch = default_git_branch

    def parse_commands(self, commands, force=False, default_git_branch=TENTACLES_DEFAULT_BRANCH):
        help_message = """- install: Install or re-install the given tentacles modules with their requirements if any.
    Does not edit tentacles configuration files and creates it if missing.
- update: Update the given tentacle modules with their requirements if any. Does not edit tentacles configuration files
- uninstall: Uninstall the given tentacle modules. Also delete the associated tentacle configuration.
- reset_tentacles: Deletes all the installed tentacle modules and configurations.
Note: install, update and uninstall commands can take 2 types of arguments:
    - all: applies the command to all available tentacles (remote and installed tentacles).
    - modules_name1, module_name2, ... : force to apply the command to the given tentacle modules 
          identified by their name and separated by a ' '."""
        self._update_git_branch(commands, default_git_branch=default_git_branch)
        self.update_list()
        if commands:
            if commands[0] == "install":
                if commands[1] == "all":
                    self.install_parser(commands, True, force=force)
                else:
                    commands.pop(0)
                    self.install_parser(commands, False, force=force)

            elif commands[0] == "update":
                if commands[1] == "all":
                    self.update_parser(commands, True)
                else:
                    commands.pop(0)
                    self.update_parser(commands, False)
                TentaclePackageManager.update_evaluator_config_file()
                TentaclePackageManager.update_trading_config_file()

            elif commands[0] == "uninstall":
                if self._confirm_action("Uninstall tentacle(s) and remove uninstalled tentacle(s) configuration ?"):
                    if commands[1] == "all":
                        self.uninstall_parser(commands, True)
                    else:
                        commands.pop(0)
                        self.uninstall_parser(commands, False)
                    TentaclePackageManager.update_evaluator_config_file()
                    TentaclePackageManager.update_trading_config_file()

            elif commands[0] == "reset_tentacles":
                if self._confirm_action("Reset ALL tentacles ? "
                                        "This will delete all tentacle files and configuration in tentacles folder."):
                    self.reset_tentacles()

            elif commands[0] in ["help", "h", "-h", "--help"]:
                self.logger.info(f"Welcome in Tentacle Manager, commands are:\n{help_message}")

            else:
                self.logger.error(f"Command not found: {commands[0]}, commands are:\n{help_message}")
        else:
            arguments_help = "-p: activates the package manager."
            self.logger.error(f"Invalid arguments, arguments are: {arguments_help}")

    def install_parser(self, commands, command_all=False, force=False):
        should_install = True
        # first ensure the current tentacles architecture is setup correctly
        if TentacleUtil.create_missing_tentacles_arch() and not force:
            should_install = self._confirm_action(self.TENTACLE_INSTALLATION_FOUND_MESSAGE)
        if should_install:
            # then process installations
            if command_all:

                self.tentacle_package_manager.set_max_steps(self._get_package_count())
                self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.INSTALL,
                                                                              self.default_package,
                                                                              EVALUATOR_DEFAULT_FOLDER)
                for package in self.advanced_package_list:
                    self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.INSTALL,
                                                                                  package, EVALUATOR_ADVANCED_FOLDER)

                nb_actions = self._get_package_count()
            else:
                nb_actions = len(commands)
                self.tentacle_package_manager.set_max_steps(nb_actions)
                for component in commands:

                    component = TentacleUtil.check_format(component)
                    package, _, localisation, is_url, destination, package_name = self.get_package_in_lists(component)

                    if package:
                        try:
                            self.tentacle_package_manager.process_module(TentacleManagerActions.INSTALL, package,
                                                                         component, localisation, is_url, destination,
                                                                         package_name)

                        except Exception as e:
                            self.logger.error(f"Installation failed for tentacle module '{component}'")
                            raise e
                    else:
                        self.logger.error(f"No installation found for tentacle module '{component}'")
                    self.tentacle_package_manager.inc_current_step()

            TentaclePackageManager.update_evaluator_config_file()
            TentaclePackageManager.update_trading_config_file()

            return nb_actions

    def update_parser(self, commands, command_all=False):
        self.tentacle_package_manager.set_installed_modules(self.tentacle_package_manager.get_installed_modules())
        if command_all:
            self.tentacle_package_manager.set_max_steps(self._get_package_count())
            self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.UPDATE,
                                                                          self.default_package,
                                                                          EVALUATOR_DEFAULT_FOLDER)
            for package in self.advanced_package_list:
                self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.UPDATE, package,
                                                                              EVALUATOR_ADVANCED_FOLDER)

            return self._get_package_count()

        else:
            nb_actions = len(commands)
            self.tentacle_package_manager.set_max_steps(nb_actions)
            for component in commands:

                component = TentacleUtil.check_format(component)
                package, _, localisation, is_url, destination, package_name = self.get_package_in_lists(component)

                if package:
                    try:
                        self.tentacle_package_manager.process_module(TentacleManagerActions.UPDATE, package, component,
                                                                     localisation, is_url, destination, package_name)

                    except Exception as e:
                        self.logger.error(f"Update failed for tentacle module '{component}'")
                        raise e
                else:
                    self.logger.error(f"No tentacle found for '{component}'")
                self.tentacle_package_manager.inc_current_step()

            return nb_actions

    def uninstall_parser(self, commands, command_all=False):
        if command_all:
            self.tentacle_package_manager.set_max_steps(self._get_package_count())
            self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.UNINSTALL,
                                                                          self.default_package,
                                                                          EVALUATOR_DEFAULT_FOLDER)
            for package in self.advanced_package_list:
                self.tentacle_package_manager.try_action_on_tentacles_package(TentacleManagerActions.UNINSTALL,
                                                                              package, EVALUATOR_ADVANCED_FOLDER)
            return self._get_package_count()
        else:
            nb_actions = len(commands)
            self.tentacle_package_manager.set_max_steps(nb_actions)
            for component in commands:

                component = TentacleUtil.check_format(component)
                package, _, _, _, destination, _ = self.get_package_in_lists(component)

                if package:
                    try:
                        self.tentacle_package_manager.process_module(TentacleManagerActions.UNINSTALL, package,
                                                                     component, "", "", destination, "")
                    except Exception:
                        self.logger.error(f"Uninstalling failed for module '{component}'")
                else:
                    self.logger.error(f"No module found for '{component}'")
                self.tentacle_package_manager.inc_current_step()
            return nb_actions

    def reset_tentacles(self):
        self.logger.info("Removing tentacles.")
        TentacleUtil.delete_tentacles_arch()
        TentacleUtil.create_missing_tentacles_arch()
        self.logger.info("Tentacles reset.")

    def update_list(self):
        default_package_list_url = TentaclePackageUtil.get_octobot_tentacle_public_repo(git_branch=self.git_branch)

        self.default_package = TentaclePackageUtil.get_package_description(default_package_list_url,
                                                                           git_branch=self.git_branch)

        if CONFIG_TENTACLES_KEY in self.config:
            for package in self.config[CONFIG_TENTACLES_KEY]:
                # try with package as in configuration
                try:
                    self.advanced_package_list.append(
                        TentaclePackageUtil.get_package_description_with_adaptation(package,
                                                                                    git_branch=self.git_branch))
                except Exception:
                    self.logger.error(f"Impossible to get an OctoBot Tentacles Package at : {package}")

    def get_package_in_lists(self, component_name, component_version=None):
        if TentacleUtil.has_required_package(self.default_package, component_name, component_version):
            package_description = self.default_package[TENTACLE_PACKAGE_DESCRIPTION]
            package_localisation = package_description[TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION]
            is_url = package_description[TENTACLE_DESCRIPTION_IS_URL]
            package_name = package_description[TENTACLE_PACKAGE_NAME]
            return self.default_package, package_description, package_localisation, is_url, EVALUATOR_DEFAULT_FOLDER, \
                package_name
        else:
            for advanced_package in self.advanced_package_list:
                if TentacleUtil.has_required_package(advanced_package, component_name, component_version):
                    package_description = advanced_package[TENTACLE_PACKAGE_DESCRIPTION]
                    package_localisation = package_description[TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION]
                    url = package_description[TENTACLE_DESCRIPTION_IS_URL]
                    package_name = package_description[TENTACLE_PACKAGE_NAME]
                    return advanced_package, package_description, package_localisation, url, \
                        EVALUATOR_ADVANCED_FOLDER, package_name
        return None, None, None, None, None, None

    def _confirm_action(self, action):
        if self.force_actions:
            return True
        else:
            confirmation = ["yes", "ye", "y", "oui", "o"]
            user_input = input(f"{action} Y/N").lower()
            if user_input in confirmation:
                return True
            else:
                self.logger.info("Action aborted.")
                return False

    def set_force_actions(self, force_actions):
        self.force_actions = force_actions

    def _get_package_count(self):
        count = len(self.default_package)
        for advanced_package in self.advanced_package_list:
            count += len(advanced_package)
        # remove 1 for tentacle description for each tentacles package
        return count - 1*(1+len(self.advanced_package_list)) + 1



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from enum import Enum
from logging import INFO

VERSION = "1.0.10"
PROJECT_NAME = "OctoBot-Tentacles-Manager"
OCTOBOT_NAME = "OctoBot"

OCTOBOT_ADV_MNG_IMPORT_CHANGE_VERSION = "0.3.4"

LOG_LEVEL = INFO

# github
GITHUB = "github"
GITHUB_RAW_CONTENT_URL = "https://raw.githubusercontent.com"
GITHUB_API_CONTENT_URL = "https://api.github.com"
GITHUB_BASE_URL = "https://github.com"
GITHUB_ORGANISATION = "Drakkar-Software"
GITHUB_REPOSITORY = f"{GITHUB_ORGANISATION}/{OCTOBOT_NAME}"
GITHUB_URL = f"{GITHUB_BASE_URL}/{GITHUB_REPOSITORY}"

CONFIG_DEBUG_OPTION = "DEV-MODE"

# Tentacles (packages)
PYTHON_INIT_FILE = "__init__.py"
CONFIG_EVALUATOR_FILE = "evaluator_config.json"
CONFIG_TRADING_FILE = "trading_config.json"
TENTACLES_PATH = "tentacles"
TENTACLES_EVALUATOR_PATH = "Evaluator"
TENTACLES_TRADING_PATH = "Trading"
CONFIG_EVALUATOR_FILE_PATH = f"{TENTACLES_PATH}/{TENTACLES_EVALUATOR_PATH}/{CONFIG_EVALUATOR_FILE}"
CONFIG_TRADING_FILE_PATH = f"{TENTACLES_PATH}/{TENTACLES_TRADING_PATH}/{CONFIG_TRADING_FILE}"
CONFIG_DEFAULT_EVALUATOR_FILE = "config/default_evaluator_config.json"
CONFIG_DEFAULT_TRADING_FILE = "config/default_trading_config.json"
TENTACLES_TEST_PATH = "tests"
TENTACLES_EVALUATOR_REALTIME_PATH = "RealTime"
TENTACLES_EVALUATOR_TA_PATH = "TA"
TENTACLES_EVALUATOR_SOCIAL_PATH = "Social"
TENTACLES_EVALUATOR_STRATEGIES_PATH = "Strategies"
TENTACLES_EVALUATOR_UTIL_PATH = "Util"
TENTACLES_TRADING_MODE_PATH = "Mode"
TENTACLES_PYTHON_INIT_CONTENT = "from .Default import *\nfrom .Advanced import *\n"
TENTACLES_PUBLIC_REPOSITORY = f"{GITHUB_ORGANISATION}/{OCTOBOT_NAME}-Tentacles"
TENTACLES_PUBLIC_LIST = "tentacles_list.json"
TENTACLES_DEFAULT_BRANCH = "master"
EVALUATOR_DEFAULT_FOLDER = "Default"
EVALUATOR_ADVANCED_FOLDER = "Advanced"
TENTACLES_INSTALL_FOLDERS = [EVALUATOR_DEFAULT_FOLDER, EVALUATOR_ADVANCED_FOLDER]
EVALUATOR_CONFIG_FOLDER = "config"
EVALUATOR_RESOURCE_FOLDER = "resources"
CONFIG_TENTACLES_KEY = "tentacles-packages"
TENTACLE_PACKAGE_DESCRIPTION = "package_description"
TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION = "localisation"
TENTACLE_PACKAGE_NAME = "name"
TENTACLE_DESCRIPTION_IS_URL = "is_url"
TENTACLE_MODULE_TESTS = "tests"
TENTACLE_MODULE_DESCRIPTION = "$tentacle_description"
TENTACLE_MODULE_REQUIREMENTS = "requirements"
TENTACLE_MODULE_REQUIREMENT_WITH_VERSION = "requirement_with_version"
TENTACLE_MODULE_LIST_SEPARATOR = ","
TENTACLE_MODULE_REQUIREMENT_VERSION_SEPARATOR = "=="
TENTACLE_MODULE_NAME = "name"
TENTACLE_MODULE_TYPE = "type"
TENTACLE_MODULE_SUBTYPE = "subtype"
TENTACLE_MODULE_VERSION = "version"
TENTACLE_MODULE_DEV = "developing"
TENTACLE_PACKAGE = "package_name"
TENTACLE_MODULE_CONFIG_FILES = "config_files"
TENTACLE_MODULE_CONFIG_SCHEMA_FILES = "config_schema_files"
TENTACLE_MODULE_RESOURCE_FILES = "resource_files"
TENTACLE_CREATOR_PATH = "tentacle_creator"
TENTACLE_TEMPLATE_DESCRIPTION = "description"
TENTACLE_TEMPLATE_PATH = "templates"
TENTACLE_TEMPLATE_PRE_EXT = "_tentacle"
TENTACLE_CONFIG_TEMPLATE_PRE_EXT = "_config"
TENTACLE_TEMPLATE_EXT = ".template"
TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION = "1.1.0"

TENTACLE_SONS = {"Social": TENTACLES_EVALUATOR_SOCIAL_PATH,
                 "RealTime": TENTACLES_EVALUATOR_REALTIME_PATH,
                 "Util": TENTACLES_EVALUATOR_UTIL_PATH,
                 "TA": TENTACLES_EVALUATOR_TA_PATH,
                 "Strategies": TENTACLES_EVALUATOR_STRATEGIES_PATH,
                 "Mode": TENTACLES_TRADING_MODE_PATH}

TENTACLE_PARENTS = {
    "Evaluator": TENTACLES_EVALUATOR_PATH,
    "Trading": TENTACLES_TRADING_PATH
}

TENTACLE_TYPES = {**TENTACLE_PARENTS, **TENTACLE_SONS}

# other
CONFIG_FILE_EXT = ".json"


class TentacleManagerActions(Enum):
    INSTALL = 1
    UNINSTALL = 2
    UPDATE = 3



#  Drakkar-Software OctoBot-Tentacles-Manager-Launcher
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

from setuptools import setup, find_packages

from tentacles_manager import PROJECT_NAME, VERSION


def find_package_data(path):
    return (path, [os.path.join(dirpath, filename)
                   for dirpath, dirnames, filenames in os.walk(path)
                   for filename in
                   [file for file in filenames if not file.endswith(".py") and not file.endswith(".pyc")]])


PACKAGES = find_packages()

PACKAGES_DATA = [find_package_data("tentacles_manager/tentacle_creator/templates")]

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()

REQUIRED = open('requirements.txt').read()
REQUIRES_PYTHON = '>=3.6'

setup(
    name=PROJECT_NAME,
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Tentacles-Manager',
    license='LGPL-3.0',
    author='Drakkar-Software',
    author_email='drakkar.software@protonmail.com',
    description='OctoBot project module installer',
    packages=PACKAGES,
    long_description=DESCRIPTION,
    install_requires=REQUIRED,
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    data_files=PACKAGES_DATA,
    python_requires=REQUIRES_PYTHON,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
    ],
)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import logging
import os
import shutil
import copy
import json

from tentacles_manager import EVALUATOR_DEFAULT_FOLDER, TENTACLE_TYPES, EVALUATOR_CONFIG_FOLDER, \
    TENTACLE_MODULE_REQUIREMENT_VERSION_SEPARATOR, TENTACLE_MODULE_NAME, TENTACLE_MODULE_TYPE, \
    TENTACLE_MODULE_SUBTYPE, TENTACLE_MODULE_VERSION, TENTACLE_MODULE_CONFIG_FILES, TENTACLE_MODULE_REQUIREMENTS, \
    TENTACLE_MODULE_LIST_SEPARATOR, TENTACLE_MODULE_REQUIREMENT_WITH_VERSION, TENTACLE_MODULE_DESCRIPTION, \
    TENTACLES_INSTALL_FOLDERS, TENTACLES_PATH, TENTACLES_EVALUATOR_PATH, TENTACLES_TRADING_PATH, \
    TENTACLES_EVALUATOR_REALTIME_PATH, TENTACLES_EVALUATOR_TA_PATH, TENTACLES_EVALUATOR_SOCIAL_PATH, \
    TENTACLES_EVALUATOR_STRATEGIES_PATH, TENTACLES_EVALUATOR_UTIL_PATH, TENTACLES_TRADING_MODE_PATH, \
    TENTACLES_PYTHON_INIT_CONTENT, PYTHON_INIT_FILE, TENTACLE_MODULE_TESTS, TENTACLES_TEST_PATH, TENTACLE_MODULE_DEV, \
    TENTACLE_PACKAGE, TENTACLE_MODULE_RESOURCE_FILES, EVALUATOR_RESOURCE_FOLDER, TENTACLE_MODULE_CONFIG_SCHEMA_FILES, \
    TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION, CONFIG_DEBUG_OPTION, INFO


def tentacles_arch_exists() -> bool:
    try:
        import tentacles
        return os.path.exists(TENTACLES_PATH) and os.path.exists(f"{TENTACLES_PATH}/{TENTACLES_TEST_PATH}")
    except ImportError:
        return False


def delete_tentacles_arch():
    if tentacles_arch_exists():
        shutil.rmtree(TENTACLES_PATH)


def check_format(component):
    purified_name = component.rstrip(",+.; ")
    purified_name = purified_name.strip(",+.; ")
    return purified_name


def has_required_package(package, component_name, component_version=None):
    if component_name in package:
        component = package[component_name]
        if component_version:
            return component[TENTACLE_MODULE_VERSION] == component_version
        else:
            return True
    return False


def create_missing_tentacles_arch():
    found_existing_installation = False
    tentacle_architecture, tentacle_extremity_architecture = get_tentacles_arch()
    for tentacle_root, subdir in tentacle_architecture.items():
        found_existing_installation = not _find_or_create(tentacle_root)
        init_path = os.path.join(tentacle_root, PYTHON_INIT_FILE)
        _find_or_create(init_path, False, "")
        for tentacle_dir in subdir:
            for tentacle_type_dir, types_subdir in tentacle_dir.items():
                type_path = os.path.join(tentacle_root, tentacle_type_dir)
                _find_or_create(type_path)
                init_path = os.path.join(type_path, PYTHON_INIT_FILE)
                _find_or_create(init_path, False, "")
                if isinstance(types_subdir, dict):
                    for tentacle_subtype_dir, types_subsubdir in types_subdir.items():
                        test_type_path = os.path.join(type_path, tentacle_subtype_dir)
                        _find_or_create(test_type_path)
                        init_path = os.path.join(test_type_path, PYTHON_INIT_FILE)
                        _find_or_create(init_path, False, "")
                        _create_arch_module_extremity(tentacle_extremity_architecture,
                                                      types_subsubdir, test_type_path, False, False)
                else:
                    _create_arch_module_extremity(tentacle_extremity_architecture, types_subdir, type_path)
    return found_existing_installation


def _create_arch_module_extremity(architecture, types_subdir, type_path, with_init_config=True, with_init_res=True):
    for module_type in types_subdir:
        path = os.path.join(type_path, module_type)
        _find_or_create(path)
        for extremity_folder in architecture:
            module_content_path = os.path.join(path, extremity_folder)
            # add Advanced etc folders
            _find_or_create(module_content_path)
            init_path = os.path.join(module_content_path, PYTHON_INIT_FILE)
            _find_or_create(init_path, False, "")
        # add init.py file
        init_path = os.path.join(path, PYTHON_INIT_FILE)
        _find_or_create(init_path, False)

        if with_init_config:
            module_config_path = os.path.join(path, EVALUATOR_CONFIG_FOLDER)
            _find_or_create(module_config_path)
            module_default_config_path = os.path.join(module_config_path, EVALUATOR_DEFAULT_FOLDER)
            _find_or_create(module_default_config_path)

        if with_init_res:
            module_res_path = os.path.join(path, EVALUATOR_RESOURCE_FOLDER)
            _find_or_create(module_res_path)


def get_tentacles_arch():
    tentacles_content_folder = {
        TENTACLES_EVALUATOR_PATH: [
            TENTACLES_EVALUATOR_REALTIME_PATH,
            TENTACLES_EVALUATOR_SOCIAL_PATH,
            TENTACLES_EVALUATOR_TA_PATH,
            TENTACLES_EVALUATOR_STRATEGIES_PATH,
            TENTACLES_EVALUATOR_UTIL_PATH
        ],
        TENTACLES_TRADING_PATH: [
            TENTACLES_TRADING_MODE_PATH
        ]
    }
    tentacle_architecture = {
        TENTACLES_PATH: [tentacles_content_folder, {TENTACLES_TEST_PATH: tentacles_content_folder}]
    }
    tentacle_extremity_architecture = copy.deepcopy(TENTACLES_INSTALL_FOLDERS)
    return tentacle_architecture, tentacle_extremity_architecture


def _find_or_create(path, is_directory=True, file_content=TENTACLES_PYTHON_INIT_CONTENT):
    if not os.path.exists(path):
        if is_directory:
            if not os.path.isdir(path):
                os.makedirs(path)
        else:
            if not os.path.isfile(path):
                # should be used for python init.py files only
                with open(path, "w+") as file:
                    file.write(file_content)
        return True
    return False


def parse_module_header(module_header_content):
    return {
        TENTACLE_MODULE_NAME: module_header_content[TENTACLE_MODULE_NAME],
        TENTACLE_MODULE_TYPE: module_header_content[TENTACLE_MODULE_TYPE],
        TENTACLE_MODULE_SUBTYPE: module_header_content[TENTACLE_MODULE_SUBTYPE]
        if TENTACLE_MODULE_SUBTYPE in module_header_content else None,
        TENTACLE_MODULE_VERSION: module_header_content[TENTACLE_MODULE_VERSION],
        TENTACLE_MODULE_REQUIREMENTS: extract_tentacle_requirements(module_header_content),
        TENTACLE_MODULE_TESTS: extract_tentacle_tests(module_header_content),
        TENTACLE_MODULE_CONFIG_FILES: module_header_content[TENTACLE_MODULE_CONFIG_FILES]
        if TENTACLE_MODULE_CONFIG_FILES in module_header_content else None,
        TENTACLE_MODULE_CONFIG_SCHEMA_FILES: module_header_content[TENTACLE_MODULE_CONFIG_SCHEMA_FILES]
        if TENTACLE_MODULE_CONFIG_SCHEMA_FILES in module_header_content else None,
        TENTACLE_MODULE_RESOURCE_FILES: module_header_content[TENTACLE_MODULE_RESOURCE_FILES]
        if TENTACLE_MODULE_RESOURCE_FILES in module_header_content else None,
        TENTACLE_MODULE_DEV: module_header_content[TENTACLE_MODULE_DEV]
        if TENTACLE_MODULE_DEV in module_header_content else None,
        TENTACLE_PACKAGE: module_header_content[TENTACLE_PACKAGE]
        if TENTACLE_PACKAGE in module_header_content else "???",
    }


def extract_tentacle_requirements(module):
    if TENTACLE_MODULE_REQUIREMENTS in module:
        requirements = []
        for component in module[TENTACLE_MODULE_REQUIREMENTS]:
            requirements = requirements + component.split(TENTACLE_MODULE_LIST_SEPARATOR)
        return [parse_requirements(req.strip()) for req in requirements]
    return None


def parse_requirements(requirement):
    requirement_info = requirement.split(TENTACLE_MODULE_REQUIREMENT_VERSION_SEPARATOR)
    return {TENTACLE_MODULE_NAME: requirement_info[0],
            TENTACLE_MODULE_VERSION: requirement_info[1] if len(requirement_info) > 1 else None,
            TENTACLE_MODULE_REQUIREMENT_WITH_VERSION: requirement}


def extract_tentacle_tests(module):
    if TENTACLE_MODULE_TESTS in module:
        tests = []
        for component in module[TENTACLE_MODULE_TESTS]:
            tests = tests + component.split(TENTACLE_MODULE_LIST_SEPARATOR)
        return [test.strip() for test in tests]
    return None


def parse_module_file(module_file_content, description_list):
    min_description_length = 10
    description_pos = module_file_content.find(TENTACLE_MODULE_DESCRIPTION)
    if description_pos > -1:
        description_begin_pos = module_file_content.find("{")
        description_end_pos = module_file_content.find("}") + 1
        if description_end_pos - description_begin_pos > min_description_length:
            description_raw = module_file_content[description_begin_pos:description_end_pos]
            description = json.loads(description_raw)
            module_description = parse_module_header(description)
            description_list[module_description[TENTACLE_MODULE_NAME]] = module_description


def create_localization_from_type(localization, module_type, module_subtype, file, tests=False):
    # create path from types
    test_folder_if_required = ""
    if tests:
        test_folder_if_required = f"/{TENTACLES_TEST_PATH}"
    if module_subtype:
        return f"{localization}{test_folder_if_required}/{module_type}/{module_subtype}/{file}"
    else:
        return f"{localization}{test_folder_if_required}/{module_type}/{file}"


def create_path_from_type(module_type, module_subtype, target_folder, tests=False):
    # create path from types
    test_folder_if_required = ""
    if tests:
        test_folder_if_required = f"/{TENTACLES_TEST_PATH}"
    if module_subtype:
        return f"{TENTACLES_PATH}{test_folder_if_required}/{TENTACLE_TYPES[module_type]}/" \
            f"{TENTACLE_TYPES[module_subtype]}/{target_folder}"
    else:
        return f"{TENTACLES_PATH}{test_folder_if_required}/{TENTACLE_TYPES[module_type]}/{target_folder}"


def get_full_module_identifier(module_name, module_version):
    return f"{module_name}{TENTACLE_MODULE_REQUIREMENT_VERSION_SEPARATOR}{module_version}"


def parse_version(version):
    return [int(value) for value in version.split(".") if value]


def is_first_version_superior(first_version, second_version, or_equal=False):
    first_version_data = parse_version(first_version)
    second_version_data = parse_version(second_version)

    if len(second_version_data) > len(first_version_data):
        return False
    else:
        if or_equal and first_version_data == second_version_data:
            return True
        for index, value in enumerate(first_version_data):
            if len(second_version_data) > index and second_version_data[index] > value:
                return False
    return first_version_data != second_version_data


def is_module_in_list(module_name, module_version, module_list):
    if not module_version:
        return any(module.split(TENTACLE_MODULE_REQUIREMENT_VERSION_SEPARATOR)[0] == module_name
                   for module in module_list)
    else:
        return get_full_module_identifier(module_name, module_version) \
               in module_list


def install_on_development(config, module_dev):
    # is not on development
    if module_dev is None or not module_dev:
        return True

    # is on development
    if module_dev and CONFIG_DEBUG_OPTION in config and config[CONFIG_DEBUG_OPTION]:
        return True

    return False


def _parse_package(tentacle_content):
    description_pos = tentacle_content.find("$tentacle_description")
    if description_pos > -1:
        description_begin_pos = tentacle_content.find("{")
        description_end_pos = tentacle_content.find("}") + 1
        description_raw = tentacle_content[description_begin_pos:description_end_pos]
        return json.loads(description_raw)
    return None


def _read_tentacle(file_name):
    if file_name.endswith(".py"):
        with open(file_name, "r") as tentacle:
            return _parse_package(tentacle.read())


def check_tentacle(path, tentacle, verbose=True):
    logger = logging.getLogger("TentacleChecker")
    logger.setLevel(INFO)
    try:
        # only check tentacle version for default tentacles
        if path.endswith(EVALUATOR_DEFAULT_FOLDER):
            tentacle_desc = _read_tentacle(os.path.join(path, f"{tentacle}.py"))
            installed_version = tentacle_desc[TENTACLE_MODULE_VERSION]
            if is_first_version_superior(installed_version, TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION,
                                         or_equal=True):
                return True
            else:
                if verbose:
                    logger.error(f"Incompatible tentacle {tentacle}: version {installed_version}, "
                                 f"minimum expected: {TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION} this tentacle "
                                 f"may not work properly. Please update your tentacles ('start.py -p update {tentacle}' "
                                 f"or 'start.py -p update all')")
                return False
    except Exception as e:
        if verbose:
            logger.error(f"Error when reading tentacle description: {e}")
        return False



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import json
import logging

import requests
import os

from tentacles_manager import tentacle_util as TentacleUtil
from tentacles_manager import TENTACLES_PUBLIC_LIST, TENTACLES_DEFAULT_BRANCH, TENTACLE_PACKAGE_DESCRIPTION, \
    GITHUB_BASE_URL, GITHUB, TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION, TENTACLE_DESCRIPTION_IS_URL, \
    TENTACLES_INSTALL_FOLDERS, PYTHON_INIT_FILE, TENTACLE_PACKAGE_NAME, TENTACLES_PUBLIC_REPOSITORY, \
    GITHUB_RAW_CONTENT_URL, INFO, OCTOBOT_ADV_MNG_IMPORT_CHANGE_VERSION


def get_package_description_with_adaptation(url_or_path, git_branch=TENTACLES_DEFAULT_BRANCH):
    # try with package as in configuration
    try:
        return get_package_description(url_or_path, git_branch=git_branch)
    except Exception:
        return get_package_description(url_or_path, True, git_branch=git_branch)


def get_package_description(url_or_path, try_to_adapt=False, git_branch=TENTACLES_DEFAULT_BRANCH):
    package_url_or_path = str(url_or_path)
    # if its an url: download with requests.get and return text
    if get_is_url(package_url_or_path):
        if try_to_adapt:
            if not package_url_or_path.endswith("/"):
                package_url_or_path += "/"
            # if checking on github, try adding branch and file
            if GITHUB in package_url_or_path:
                package_url_or_path += f"{git_branch}/{TENTACLES_PUBLIC_LIST}"
            # else try adding file
            else:
                package_url_or_path += TENTACLES_PUBLIC_LIST
        downloaded_result = json.loads(requests.get(package_url_or_path).text)
        if "error" in downloaded_result and GITHUB_BASE_URL in package_url_or_path:
            package_url_or_path = package_url_or_path.replace(GITHUB_BASE_URL, GITHUB_RAW_CONTENT_URL)
            downloaded_result = json.loads(requests.get(package_url_or_path).text)
        # add package metadata
        add_package_description_metadata(downloaded_result, package_url_or_path, True)
        return downloaded_result

    # if its a local path: return file content
    else:
        if try_to_adapt:
            if not package_url_or_path.endswith("/"):
                package_url_or_path += "/"
            package_url_or_path += TENTACLES_PUBLIC_LIST
        with open(package_url_or_path, "r") as package_description:
            read_result = json.loads(package_description.read())
            # add package metadata
            add_package_description_metadata(read_result, package_url_or_path, False)
            return read_result


def get_is_url(string):
    return string.startswith("https://") \
        or string.startswith("http://") \
        or string.startswith("ftp://")


def add_package_description_metadata(package_description, localisation, is_url):
    to_save_loc = str(localisation)
    if localisation.endswith(TENTACLES_PUBLIC_LIST):
        to_save_loc = localisation[0:-len(TENTACLES_PUBLIC_LIST)]
        while to_save_loc.endswith("/") or to_save_loc.endswith("\\"):
            to_save_loc = to_save_loc[0:-1]
    package_description[TENTACLE_PACKAGE_DESCRIPTION] = {
        TENTACLE_PACKAGE_DESCRIPTION_LOCALISATION: to_save_loc,
        TENTACLE_PACKAGE_NAME: get_package_name(localisation, is_url),
        TENTACLE_DESCRIPTION_IS_URL: is_url
    }


def get_octobot_tentacle_public_repo(to_description_file=True, git_branch=TENTACLES_DEFAULT_BRANCH):
    if to_description_file:
        return f"{GITHUB_BASE_URL}/{TENTACLES_PUBLIC_REPOSITORY}/{git_branch}/{TENTACLES_PUBLIC_LIST}"
    else:
        return f"{GITHUB_BASE_URL}/{TENTACLES_PUBLIC_REPOSITORY}"


def add_package_name(module_file_content, package_name):
    return module_file_content.replace('$tentacle_description: {\n', '$tentacle_description: {\n    ' +
                                       f'"package_name": "{package_name}",\n')


def get_package_name(localisation, is_url):
    if is_url:
        github_package_name_reverse_index = -3 if localisation.endswith(".json") else -1
        return localisation.split("/")[github_package_name_reverse_index]
    else:
        local_package_name_reverse_index = -2 if localisation.endswith(".json") else -1
        separator = "/"
        if "\\" in localisation:
            separator = "\\"
        return localisation.split(separator)[local_package_name_reverse_index]


def get_package_file_content_from_url(url, as_bytes=False):
    result = requests.get(url)
    package_file = result.text

    if package_file.find("404: Not Found") != -1:
        raise Exception(f"{package_file}: {url}")

    if as_bytes:
        return result.content
    else:
        return package_file


def read_tentacles(path, description_list):
    for file_name in os.listdir(path):
        if check_path(path) and file_name.endswith(".py") \
                and file_name != PYTHON_INIT_FILE:
            with open(f"{path}/{file_name}", "r") as module:
                TentacleUtil.parse_module_file(module.read(), description_list)
        else:
            file_name = f"{path}/{file_name}"
            if os.path.isdir(file_name) and not path.startswith('.'):
                read_tentacles(file_name, description_list)


def should_recreate_config_file(file_path, config_file_content):
    return config_file_content and not os.path.isfile(file_path)


def check_path(path):
    last_path_folder = path.split("/")[-1]
    return last_path_folder in TENTACLES_INSTALL_FOLDERS


def update_config_file(config_file_path, default_file_path, classes_to_consider):

    logger = logging.getLogger("TentaclePackageUtil")
    logger.setLevel(INFO)

    # initialize file content
    config_content = {}
    changed_something = False
    if os.path.isfile(config_file_path):
        with open(config_file_path, "r") as config_file_r:
            default_config_file_content = config_file_r.read()
            try:
                config_content = json.loads(default_config_file_content)
            except Exception as e:
                logger.warning(f"impossible to load content of configuration file: "
                               f"{config_file_path}: {e}")
    # take default values into account using default file
    if os.path.isfile(default_file_path):
        with open(default_file_path, "r") as default_config_file_r:
            default_config_file_content = default_config_file_r.read()
            default_config_content = json.loads(default_config_file_content)
            for key, val in default_config_content.items():
                if key not in config_content:
                    config_content[key] = val
                    changed_something = True

    classes_list = []
    # add items using their base class key (vs advances classes)
    for classes_in_config in classes_to_consider:
        changed_something = add_class_to_config_file_content(
            classes_in_config, config_content, classes_list) or changed_something

    # remove potential unnecessary items
    to_remove = []
    str_classes_list = [c.get_name() for c in classes_list]
    for key in config_content.keys():
        if key not in str_classes_list:
            to_remove.append(key)

    for key_to_remove in to_remove:
        config_content.pop(key_to_remove)
        changed_something = True

    if changed_something:
        with open(config_file_path, "w+") as config_file_w:
            config_file_w.write(json.dumps(config_content, indent=4, sort_keys=True))
            logger.info(f"{config_file_path} has been updated")
    else:
        logger.info(f"Nothing to update in {config_file_path}")


def add_class_to_config_file_content(clazz, config_file_content, classes_list, activated=False):
    from config import SHORT_VERSION
    if TentacleUtil.is_first_version_superior(SHORT_VERSION, OCTOBOT_ADV_MNG_IMPORT_CHANGE_VERSION, or_equal=True):
        from tentacles_management.advanced_manager import AdvancedManager
    else:
        from evaluator.Util.advanced_manager import AdvancedManager
    changed_something = False
    current_classes_list = AdvancedManager.create_default_types_list(clazz)
    for current_class in current_classes_list:
        if not current_class.get_name() in config_file_content:
            config_file_content[current_class.get_name()] = activated
            changed_something = True
    classes_list += current_classes_list
    return changed_something
